//************************************************************************
// Example9PieceTour.java
//
// Our starting point is Example8KnightTour.java, where we constructed
// knight tours of chessboards using a greedy algorithm. Our purpose
// here is to demonstrate simple examples of inheritance. Class King
// inherits from Knight; basically, the only difference is that the
// steps a king can take are different than those a knight can. We
// added a rook also, which caused more problems since the possible
// moves depend on the size of the board as well. We have changed
// class Board to deal with general rectangles, not necessarily
// square boards. Class PuncturedBoard inherits from Board and enables
// us to deal with boards with some (rectangle) hole in it.
// In class Board we also added static variables.
//************************************************************************

class Example9PieceTour
{
   public static void main(String[] args)
   {
      final int SIZE_X = Integer.parseInt(args[0]),
                SIZE_Y = Integer.parseInt(args[1]);
      for (int i=1; i <= 100; i++)
      {
         Board myBoard = new Board (SIZE_X, SIZE_Y);
         int x = (int) (myBoard.getSizeX() * Math.random());
         int y = (int) (myBoard.getSizeY() * Math.random());
         Point location = new Point (x, y);
         Knight myKnight = new Knight (location, myBoard);
         while (myKnight.canMove (myBoard))
            myKnight.move (myBoard);
      }
      System.out.println ("Tried: " + Board.getNumberOfBoards() + "\n" +
                          "Succeeded: " + Board.getNumberOfSuccesses() + "\n");

      Board myBoard = new Board (SIZE_X, SIZE_Y);
      int x = (int) (myBoard.getSizeX() * Math.random());
      int y = (int) (myBoard.getSizeY() * Math.random());
      Point location = new Point (x, y);
      King myKing = new King (location, myBoard);
      while (myKing.canMove (myBoard))
         myKing.move (myBoard);
      System.out.println (myBoard);

      myBoard = new Board (SIZE_X, SIZE_Y);
      Rook myRook = new Rook (new Point (2, 7), myBoard);
      while (myRook.canMove (myBoard))
         myRook.move (myBoard);
      System.out.println (myBoard);

      PuncturedBoard myPuncturedBoard = new PuncturedBoard (3, 3, 3, 3, 3, 3);
      Knight myKnight = new Knight (new Point (2, 7), myPuncturedBoard);
      while (myKnight.canMove (myPuncturedBoard))
         myKnight.move (myPuncturedBoard);
      System.out.println (myPuncturedBoard);
   }// method main
}// class Example9PieceTour


class Knight
{
   //===========================================================
   // Supports a few actions of the knight.
   //===========================================================
   static final int [][] STEPS = {{2, 1}, {1, 2}, {-1, 2}, {-2, 1},
                                  {-2, -1}, {-1, -2}, {1, -2}, {2, -1}};
   private Point position;

   Knight (Point location, Board myBoard)
   {placeAt (location, myBoard);}

   int[][] getSteps(Board myBoard)
   {return STEPS;}

   private void placeAt (Point location, Board myBoard) // place the knight
   {
      position = new Point (location);
      myBoard.placeKnight (position);
   }// method placeAt

   boolean canMove (Board myBoard) // check if the knight can move at all
   {return position.numberOfExits (this, myBoard) > 0;}

   // Moves the knight. First we check all possible continuations to see
   // what is the minimal number of exits from each of them. Then we find
   // all moves leading to a point with such a number of exits. Finally,
   // we selct randomly one of these points.
   void move (Board myBoard) 
   {
      int minExits = minNumberOfExits (myBoard);
      int possibilities = numberOfHardPoints (minExits, myBoard);
      int which = (int) (possibilities * Math.random() + 1);
      Point next = newLocation (which, myBoard, minExits);
      placeAt (next, myBoard);
   }// method move

   // Finds minimal number of exits from any of the points the knight can
   // move to.
   private int minNumberOfExits (Board myBoard)
   {
      int exits = getSteps(myBoard).length + 1;
      for (int i=0; i < getSteps(myBoard).length; i++)
      {
         Point next = position.translate (getSteps(myBoard)[i]);
         if (next.inBoard (myBoard) && !next.visited (myBoard))
            exits = Math.min (exits, next.numberOfExits (this, myBoard));
      }
      return exits;
   }// method minNumberOfExits

   // Finds the number of points with a minimal possible number of
   // continuations at the next stage.
   private int numberOfHardPoints (int exits, Board myBoard)
   {
      int number = 0;
      for (int i=0; i < getSteps(myBoard).length; i++)
      {
         Point next = position.translate (getSteps(myBoard)[i]);
         if (next.inBoard (myBoard) && !next.visited (myBoard)
                && next.numberOfExits (this, myBoard) == exits)
            number++;
      }
      return number;
   }// method numberOfHardPoints

   // Selects the point the knight moves to.
   private Point newLocation (int which, Board myBoard, int exits)
   {
      int number = 0, i = 0;
      Point next = new Point();
      while (number < which)
      {
         next = position.translate (getSteps(myBoard)[i]);
         if (next.inBoard (myBoard) && !next.visited (myBoard)
                && next.numberOfExits (this, myBoard) == exits)
            number++;
         i++;
      }
      return next;
   }// method newLocation
}// class Knight


class King extends Knight
{
   //===========================================================
   // Supports a few actions of the king.
   //===========================================================
   static final int[][] KING_STEPS = {{1, 0}, {1, 1}, {0, 1}, {-1, 1},
                                      {-1, 0}, {-1, -1}, {0, -1}, {1, -1}};

   King (Point location, Board myBoard)
   {super (location, myBoard);}

   int[][] getSteps(Board myBoard)
   {return KING_STEPS;}
}// class King


class Rook extends Knight
{
   //===========================================================
   // Supports a few actions of the rook.
   //===========================================================
   final int [][] ROOK_STEPS;

   Rook (Point location, Board myBoard)
   {
      super (location, myBoard);
      final int MAX_MOVES = 2 * myBoard.getSizeX() +
                            2 * myBoard.getSizeY() - 4;
      ROOK_STEPS = new int [MAX_MOVES] [2];
      for (int i=0; i <= myBoard.getSizeX() - 2; i++)
      {
         ROOK_STEPS [i][0] = i + 1;
         ROOK_STEPS [i + myBoard.getSizeX() - 1] [0] = -(i + 1);
      }
      for (int i = 2 * myBoard.getSizeX() - 2;
               i <= 2 * myBoard.getSizeX() + myBoard.getSizeY() - 4; i++)
      {
         ROOK_STEPS [i][1] = i - 2 * myBoard.getSizeX() + 3;
         ROOK_STEPS [i + myBoard.getSizeY() - 1] [1]
            = -(i - 2 * myBoard.getSizeX() + 3);
      }
   }

   int[][] getSteps (Board myBoard)
   {return ROOK_STEPS;}
}// class Rook


class Board
{
   //===========================================================
   // Supports a few actions on the board.
   //===========================================================
   final int SIZE_X, SIZE_Y;
   int[][] timeOfVisit;
   int numberOfVisitedPoints = 0;
   static int numberOfBoards = 0, numberOfSuccesses = 0;

   Board (int SIZE_X, int SIZE_Y)
   {
      this.SIZE_X = SIZE_X;
      this.SIZE_Y = SIZE_Y;
      timeOfVisit = new int [SIZE_X] [SIZE_Y];
      for (int i=0; i < SIZE_X; i++)
         for (int j=0; j < SIZE_Y; j++)
            timeOfVisit [i][j] = 0;
      numberOfBoards++;
   }// method Board

   int getSizeX()
   {return SIZE_X;}

   int getSizeY()
   {return SIZE_Y;}

   int getTimeOfVisit (int x, int y)
   {return timeOfVisit[x][y];}

   static int getNumberOfBoards()
   {return numberOfBoards;}

   static int getNumberOfSuccesses()
   {return numberOfSuccesses;}

   void placeKnight (Point P) // change board information as knight moves
   {
      numberOfVisitedPoints++;
      if (numberOfVisitedPoints == getSizeX() * getSizeY())
         numberOfSuccesses++;
      timeOfVisit [P.getX()] [P.getY()] = numberOfVisitedPoints;
   }// method placeKnight

   public String toString()
   {
      String result = "";
      for (int row=SIZE_Y-1; row >= 0; row--)
      {
         for (int col=0; col < SIZE_X; col++)
         {
            if (timeOfVisit [col][row] < 10)
               result += " ";
            if (timeOfVisit [col][row] == -1)
               result += "- ";
            else
               result += timeOfVisit [col][row] + " ";
         }
         result += "\n";
      }
      return result;
   }// method toString
}// class Board


class PuncturedBoard extends Board
{
   //===========================================================
   // Designed to deal with boards with one hole inside.
   //===========================================================
   PuncturedBoard (int SIZE_X_LEFT, int SIZE_X_HOLE, int SIZE_X_RIGHT,
                   int SIZE_Y_LEFT, int SIZE_Y_HOLE, int SIZE_Y_RIGHT)
   {
      super (SIZE_X_LEFT + SIZE_X_HOLE + SIZE_X_RIGHT,
             SIZE_Y_LEFT + SIZE_Y_HOLE + SIZE_Y_RIGHT);
      for (int i=SIZE_X_LEFT; i < SIZE_X_LEFT + SIZE_X_HOLE; i++)
         for (int j=SIZE_Y_LEFT; j < SIZE_Y_LEFT + SIZE_Y_HOLE; j++)
            timeOfVisit [i][j] = -1;
   }// method PuncturedBoard
}// class PuncturedBoard


class Point
{
   //===========================================================
   // Supports actions on squares in the board.
   //===========================================================
   private int x, y;

   Point ()
   {}

   Point (int x, int y)
   {
      this.x = x;
      this.y = y;
   }// method Point

   Point (Point P)
   {this (P.x, P.y);}

   int getX()
   {return x;}

   int getY()
   {return y;}

   Point translate (int[] by)
   {return new Point (x + by[0], y + by[1]);}

   boolean inBoard (Board myBoard) // checks if point lies within the board
   {return 0 <= x & x < myBoard.getSizeX() & 0 <= y & y < myBoard.getSizeY();}

   boolean visited (Board myBoard) // checks if point has been visited earlier
   {return myBoard.getTimeOfVisit(x, y) != 0;}

   // Counts number of moves a knight could make from this point given
   // current board situation.
   int numberOfExits (Knight myKnight, Board myBoard)
   {
      int result = 0;
      for (int i=0; i < myKnight.getSteps(myBoard).length; i++)
      {
         Point next = translate (myKnight.getSteps(myBoard)[i]);
         if (next.inBoard (myBoard) && !next.visited (myBoard))
            result++;
      }
      return result;
   }// method numberOfExits
}// class Point
